import os
import socket


class Configuration(object):
    """Parameters that are used across clients and error notices

    Parameters
    ----------
    api_key : str
        errbit provided to the app to associate errors with it
    errbit_url : str
        protocol and host of the errbit server (e.g. "https://api.airbrake.io")
    project_root : str, optional
        Directory of the project's source code which is replaced by
        [PROJECT_ROOT] in the backtrace and linked to line on github.
        (the defaults is the current working directory)
    environment_name : str
        Environment field on the error's page (the default is 'production')
    server_name : str
        App Server field on error's page (the default is the server's hostname)
    """

    def __init__(self, api_key, errbit_url, project_root=None, environment_name='production', server_name=None):
        self.api_key = api_key
        self.errbit_url = errbit_url

        self.notifier_name = 'Python Errbit Reporter'
        self.notifier_version = '0.0.1'
        self.notifier_url = 'https://github.com/dylanahsmith/python-errbit-reporter'

        self.project_root = project_root or os.getcwd()
        self.environment_name = environment_name
        self.server_name = server_name or socket.gethostname()



import sys
import types
import traceback
import re
from xml.etree import cElementTree as ET

import six


class Notice(object):
    INVALID_TAG_CHARS = re.compile("[^a-zA-Z0-9_-]")

    def __init__(self, config, error_class, error_message, backtrace=None):
        self.config = config

        if isinstance(error_class, type):
            error_class = error_class.__name__
        self.error_class = error_class
        if isinstance(error_message, Exception):
            error_message = error_class + ": " + str(error_message)
        self.error_message = error_message
        if isinstance(backtrace, types.TracebackType):
            backtrace = traceback.extract_tb(backtrace)
        elif backtrace is None:
            backtrace = sys.extract_stack()
        self.backtrace = backtrace

        self.request_url = None
        self.component = None
        self.action = None
        self.params = {}
        self.session = {}
        self.cgi_data = {}

    @classmethod
    def from_exception(cls, config, exc_info=None):
        if exc_info is None:
            exc_info = sys.exc_info()
        exc_type, exc_value, exc_tb = exc_info
        return cls(config, *exc_info)

    def serialize(self):
        root = ET.Element('notice', version="2.4")

        ET.SubElement(root, 'api-key').text = self.config.api_key

        notifier = ET.SubElement(root, 'notifier')
        ET.SubElement(notifier, 'name').text = self.config.notifier_name
        ET.SubElement(notifier, 'version').text = self.config.notifier_version
        ET.SubElement(notifier, 'url').text = self.config.notifier_url

        error = ET.SubElement(root, 'error')
        ET.SubElement(error, 'class').text = self.error_class
        ET.SubElement(error, 'message').text = self.error_message
        backtrace = ET.SubElement(error, 'backtrace')
        for filename, line_number, function_name, text in reversed(self.backtrace):
            ET.SubElement(
                backtrace, 'line', number=str(line_number), file=filename, method=function_name)

        request = ET.SubElement(root, 'request')
        ET.SubElement(request, 'url').text = self.request_url
        ET.SubElement(request, 'component').text = self.component
        ET.SubElement(request, 'action').text = self.action
        self._add_xml_dict(ET.SubElement(request, 'params'), self.params)
        self._add_xml_dict(ET.SubElement(request, 'session'), self.session)
        self._add_xml_dict(ET.SubElement(request, 'cgi-data'), self.cgi_data)

        server_env = ET.SubElement(root, 'server-environment')
        ET.SubElement(
            server_env, 'project-root').text = self.config.project_root
        ET.SubElement(
            server_env, 'environment-name').text = self.config.environment_name
        ET.SubElement(server_env, 'hostname').text = self.config.server_name

        buffer = six.BytesIO()
        ET.ElementTree(root).write(
            buffer, encoding='utf-8', xml_declaration=True)
        xml_string = buffer.getvalue()
        buffer.close()
        return xml_string

    def _sanitize_tag(self, tag):
        "Sanitize tag string to avoid xml injection from user supplied params"
        if not tag:
            tag = "_"
        tag = self.INVALID_TAG_CHARS.sub('-', tag)
        if tag[0] == "-":
            tag[0] = "_"
        return tag

    def _add_xml_dict(self, parent, args):
        for name, value in six.iteritems(args):
            child = ET.SubElement(parent, name)
            self._add_xml_value(child, value)

    def _add_xml_value(self, parent, value):
        if isinstance(value, dict):
            self._add_xml_dict(parent, value)
        elif isinstance(value, list):
            for item in value:
                child = ET.SubElement(parent, 'item')
                self._add_xml_value(child, item)
        else:
            parent.text = str(value)


class Metadata(object):
    def __init__(self, id, err_id, problem_id, app_id, created_at, updated_at):
        self.id = id
        self.err_id = err_id
        self.problem_id = problem_id
        self.app_id = app_id
        self.created_at = created_at
        self.updated_at = updated_at

    @classmethod
    def from_notice_xml(cls, body):
        tree = ET.fromstring(body)

        id = tree.findtext('./_id')
        err_id = tree.findtext('./err-id')
        problem_id = tree.findtext('./problem-id')
        app_id = tree.findtext('./app-id')
        created_at = tree.findtext('./created-at')
        updated_at = tree.findtext('./updated-at')

        return cls(id, err_id, problem_id, app_id, created_at, updated_at)



from six.moves import urllib

from errbit_reporter.notice import Notice, Metadata


class Client(object):

    def __init__(self, config):
        self.config = config

    def notify(self, exc_info=None, request_url=None, component=None,
               action=None, params={}, session={}, cgi_data={}, timeout=None):
        notice = Notice.from_exception(self.config, exc_info)
        notice.request_url = request_url
        notice.component = component
        notice.action = action
        notice.params = params
        notice.session = session
        notice.cgi_data = cgi_data
        return self.send_notice(notice, timeout=timeout)

    def send_notice(self, notice, timeout=None):
        url = urllib.parse.urljoin(self.config.errbit_url, "/notifier_api/v2/notices/")
        request = urllib.request.Request(url, notice.serialize())
        request.add_header('Content-Type', 'text/xml')
        request.add_header('Accept', 'text/xml, application/xml')
        response = urllib.request.urlopen(request, timeout=timeout)
        return Metadata.from_notice_xml(response.read())



from setuptools import setup
from errbit_reporter.version import VERSION


setup(name='errbit-reporter',
      version=VERSION,
      description='Errbit Client',
      author='Dylan Thacker-Smith',
      author_email='Dylan.Smith@shopify.com',
      url='https://github.com/dylanahsmith/errbit-reporter-python',
      packages=['errbit_reporter'],
      license='MIT License',
      install_requires=[
          'six',
      ],
      test_suite='test',
      platforms='Any',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.4',
          'Topic :: Software Development :: Libraries :: Python Modules'
      ])
